print("import: 'bin'")
import bin

print("import: 'theano'")
import theano

print("import: 'theano.compile'")
import theano.compile

print("import: 'theano.compile.sandbox'")
import theano.compile.sandbox

print("import: 'theano.compile.tests'")
import theano.compile.tests

print("import: 'theano.gof'")
import theano.gof

print("import: 'theano.gof.tests'")
import theano.gof.tests

print("import: 'theano.gpuarray'")
import theano.gpuarray

print("import: 'theano.gpuarray.dnn'")
import theano.gpuarray.dnn

print("import: 'theano.gpuarray.linalg'")
import theano.gpuarray.linalg

print("import: 'theano.misc'")
import theano.misc

print("import: 'theano.sandbox'")
import theano.sandbox

print("import: 'theano.scalar'")
import theano.scalar

print("import: 'theano.scalar.tests'")
import theano.scalar.tests

print("import: 'theano.sparse'")
import theano.sparse

print("import: 'theano.sparse.tests'")
import theano.sparse.tests

print("import: 'theano.tensor'")
import theano.tensor

print("import: 'theano.tensor.nnet'")
import theano.tensor.nnet

print("import: 'theano.tensor.nnet.tests'")
import theano.tensor.nnet.tests

print("import: 'theano.tensor.signal'")
import theano.tensor.signal

print("import: 'theano.tensor.signal.tests'")
import theano.tensor.signal.tests

print("import: 'theano.tensor.tests'")
import theano.tensor.tests

print("import: 'theano.tests'")
import theano.tests

